// Copyright 2015, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Author: Zhiwei Zhang <zhiwei.zhang@saturn.uni-freiburg.de>

#include <assert.h>
#include <random>
#include <vector>
#include "./SimulationTestClass.h"
#include "./source/GenerateNumber.h"
#include "./source/Other/EliasDeltaEncodingVector.h"
#include "./source/Other/EliasDeltaEncodingPointer.h"
#include "./source/Other/EliasGammaEncodingVector.h"
#include "./source/Other/EliasGammaEncodingPointer.h"
#include "./source/Other/GolombEncodingVector.h"
#include "./source/Other/GolombEncodingPointer.h"
#include "./source/Simple8b/Simple8bCode.h"
#include "./source/Simple8b/Simple8bCodeOld.h"
#include "./source/Simple8b/Simple8bCodeVector.h"
#include "./source/Simple8b/Simple8bCodeNew.h"
#include "./source/Other/VariableByteEncodingVector.h"
#include "./source/Other/VariableByteEncodingPointer.h"

using std::vector;

// _____________________________________________________________________________
void SimulationTestClass::testEliasGammaEncodingVector() const {
  for (size_t i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      vector<size_t> plaintext;
      Bitvector encoded;
      vector<size_t> decoded;
      GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
      EliasGammaEncodingVector::encode(plaintext, encoded, true);
      EliasGammaEncodingVector::decode(encoded, decoded, true);
      for (size_t j = 0;  j < i; j++) {
        assert(decoded[j] == plaintext[j]);
      }
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testEliasGammaEncodingPointer() const {
  for (size_t i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      size_t* plaintext = new size_t[i];
      bool* encoded = new bool[i * 100];
      size_t* decoded = new size_t[i];
      GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
      EliasGammaEncodingPointer::encode(plaintext, i, encoded, true);
      EliasGammaEncodingPointer::decode(encoded, i, decoded, true);
      for (size_t j = 0;  j < i; j++) {
        assert(decoded[j] == plaintext[j]);
      }
      delete[] plaintext;
      delete[] encoded;
      delete[] decoded;
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testEliasDeltaEncodingVector() const {
  for (size_t i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      vector<size_t> plaintext;
      Bitvector encoded;
      vector<size_t> decoded;
      GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
      EliasDeltaEncodingVector::encode(plaintext, encoded, true);
      EliasDeltaEncodingVector::decode(encoded, decoded, true);
      for (size_t j = 0;  j < i; j++) {
        assert(decoded[j] == plaintext[j]);
      }
    }
  }
}
// _____________________________________________________________________________
void SimulationTestClass::testEliasDeltaEncodingPointer() const {
  for (size_t i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      size_t* plaintext = new size_t[i];
      bool* encoded = new bool[i * 50];
      size_t* decoded = new size_t[i];
      GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
      EliasDeltaEncodingPointer::encode(plaintext, i, encoded, true);
      EliasDeltaEncodingPointer::decode(encoded, i, decoded, true);
      for (size_t j = 0;  j < i; j++) {
        assert(decoded[j] == plaintext[j]);
      }
      delete[] plaintext;
      delete[] encoded;
      delete[] decoded;
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testGolombEncodingVector(size_t M) const {
  for (size_t i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      vector<int> plaintext;
      Bitvector encoded;
      vector<int> decoded;
      GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
      GolombEncodingVector::encode(plaintext, M, encoded, true);
      GolombEncodingVector::decode(encoded, M, decoded, true);
      for (size_t j = 0;  j < i; j++) {
        assert(decoded[j] == plaintext[j]);
      }
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testGolombEncodingPointer(size_t M) const {
  for (size_t i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      int* plaintext = new int[i];
      bool* encoded = new bool[i * 50];
      int* decoded = new int[i];
      GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
      GolombEncodingPointer::encode(plaintext, M, i, encoded, true);
      GolombEncodingPointer::decode(encoded, M, i, decoded, true);
      for (size_t j = 0;  j < i; j++) {
        assert(decoded[j] == plaintext[j]);
      }
      delete[] plaintext;
      delete[] encoded;
      delete[] decoded;
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testVariableByteEncodingVector() const {
  for (size_t i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      // for VB we must use the bigger type such that the errors don't come out.
      vector<int64_t> plaintext;
      VariableByteEncodingVector::Bytevector encoded;
      vector<int64_t> decoded;
      if (_geo == true) {
        GenerateNumber::geometricDistribution(0.6, i, plaintext);
      } else {
        if (_min == 0) {
          GenerateNumber::mix(_min, _max, 0.6, i, plaintext);
        } else {
          GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
        }
      }
      VariableByteEncodingVector::encode(plaintext, encoded, true);
      VariableByteEncodingVector::decode(encoded, decoded, true);
      for (size_t j = 0;  j < i; ++j) {
        assert(decoded[j] == plaintext[j]);
      }
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testVariableByteEncodingPointer() const {
  for (size_t i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      // for VB we must use the bigger type such that the errors don't come out.
      int64_t* plaintext = new int64_t[i];
      char* encoded = new char[i * 5];
      int64_t* decoded = new int64_t[i];
      if (_geo == true) {
        GenerateNumber::geometricDistribution(0.6, i, plaintext);
      } else {
        if (_min == 0) {
          GenerateNumber::mix(_min, _max, 0.6, i, plaintext);
        } else {
          GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
        }
      }
      VariableByteEncodingPointer::encode(plaintext, i, encoded, true);
      VariableByteEncodingPointer::decode(encoded, i, decoded, true);
      for (size_t j = 0;  j < i; ++j) {
        assert(decoded[j] == plaintext[j]);
      }
      delete[] plaintext;
      delete[] encoded;
      delete[] decoded;
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testSimple8bCode() const {
  for (int i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      int* plaintext = new int[i];
      uint64_t* encoded = new uint64_t[i];
      int* decoded = new int[i + 239];
      if (_geo == true) {
        GenerateNumber::geometricDistribution(0.6, i, plaintext);
      } else {
        if (_min == 0) {
          GenerateNumber::mix(_min, _max, 0.6, i, plaintext);
        } else {
          GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
        }
      }
      ad_utility::Simple8bCode::encode(plaintext, i, encoded, true);
      ad_utility::Simple8bCode::decode(encoded, i, decoded, true);
      for (int j = 0; j < i; ++j) {
        assert(decoded[j] == plaintext[j]);
      }
      delete[] plaintext;
      delete[] encoded;
      delete[] decoded;
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testSimple8bCodeOld() const {
  for (int i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      int* plaintext = new int[i];
      uint64_t* encoded = new uint64_t[i];
      int* decoded = new int[i + 239];
      if (_geo == true) {
        GenerateNumber::geometricDistribution(0.6, i, plaintext);
      } else {
        if (_min == 0) {
          GenerateNumber::mix(_min, _max, 0.6, i, plaintext);
        } else {
          GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
        }
      }
      ad_utility_old::Simple8bCode::encode(plaintext, i, encoded, true);
      ad_utility_old::Simple8bCode::decode(encoded, i, decoded, true);
      for (int j = 0; j < i; ++j) {
        assert(decoded[j] == plaintext[j]);
      }
      delete[] plaintext;
      delete[] encoded;
      delete[] decoded;
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testSimple8bCodeVector() const {
  for (int i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      vector<int> plaintext;
      vector<uint64_t> encoded;
      vector<int> decoded;
      plaintext.reserve(i);
      encoded.reserve(i);
      decoded.reserve(i);
      if (_geo == true) {
        GenerateNumber::geometricDistribution(0.6, i, plaintext);
      } else {
        if (_min == 0) {
          GenerateNumber::mix(_min, _max, 0.6, i, plaintext);
        } else {
          GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
        }
      }
      ad_utility_vector::Simple8bCodeVector::encode(
                                                  plaintext, i, encoded, true);
      ad_utility_vector::Simple8bCodeVector::decode(encoded, i, decoded, true);
      for (int j = 0; j < i; ++j) {
        assert(decoded[j] == plaintext[j]);
      }
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testSimple8bCodeNew1() const {
  for (int i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      int* plaintext = new int[i];
      uint64_t* encoded = new uint64_t[i];
      int* decoded = new int[i + 239];
      if (_geo == true) {
        GenerateNumber::geometricDistribution(0.6, i, plaintext);
      } else {
        if (_min == 0) {
          GenerateNumber::mix(_min, _max, 0.6, i, plaintext);
        } else {
          GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
        }
      }
      ad_utility_new::Simple8bCodeNew1::encode(plaintext, i, encoded, true);
      ad_utility_new::Simple8bCodeNew1::decode(encoded, i, decoded, true);
      for (int j = 0; j < i; ++j) {
        assert(decoded[j] == plaintext[j]);
      }
      delete[] plaintext;
      delete[] encoded;
      delete[] decoded;
    }
  }
}

// _____________________________________________________________________________
void SimulationTestClass::testSimple8bCodeNew2() const {
  for (int i = 1000; i <= _quantity * 10000; i *= 2) {
    for (int m = 0;  m < 10; m++) {
      int* plaintext = new int[i];
      uint64_t* encoded = new uint64_t[i];
      int* decoded = new int[i + 239];
      if (_geo == true) {
        GenerateNumber::geometricDistribution(0.6, i, plaintext);
      } else {
        if (_min == 0) {
          GenerateNumber::mix(_min, _max, 0.6, i, plaintext);
        } else {
          GenerateNumber::uniformDistribution(_min, _max, i, plaintext);
        }
      }
      ad_utility_new::Simple8bCodeNew2::encode(plaintext, i, encoded, true);
      ad_utility_new::Simple8bCodeNew2::decode(encoded, i, decoded, true);
      for (int j = 0; j < i; ++j) {
        assert(decoded[j] == plaintext[j]);
      }
      delete[] plaintext;
      delete[] encoded;
      delete[] decoded;
    }
  }
}
