// Copyright 2015, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Author: Zhiwei Zhang <zhiwei.zhang@saturn.uni-freiburg.de>

#ifndef  ZHIWEI_ZHANG_SOURCE_GENERATENUMBER_H_
#define  ZHIWEI_ZHANG_SOURCE_GENERATENUMBER_H_

#include <random>
#include <stdio.h>
#include <iostream>
#include <vector>

using std::vector;

class GenerateNumber {
 public:

  // generate the numbers which have the equal probablity.
  // this is for Simple8bCode.
  template<typename Numeric>
  static void uniformDistribution(const int& min, const int& max,
                           const size_t& quantity, Numeric* result) {
    std::random_device rd;
    std::mt19937 gen(rd());

    // make ture the numbers come out of the domain [min, max].
    std::uniform_int_distribution<> dis(min, max);

    // generate the numbers according to the given quantity.
    for (size_t n = 0; n < quantity; ++n) {
      result[n] = dis(gen);
    }
  }

  // generate the numbers which have the equal probablity.
  // this is for others.
  template<typename Numeric>
  static void uniformDistribution(const int& min, const int& max,
                  const size_t& quantity, vector<Numeric>& result) {
    result.reserve(quantity);
    std::random_device rd;
    std::mt19937 gen(rd());

    // make ture the numbers come out of the domain [min, max].
    std::uniform_int_distribution<> dis(min, max);

    // generate the numbers according to the given quantity.
    for (size_t n = 0; n < quantity; ++n) {
      result.push_back(dis(gen));
    }
  }

  // generate the numbers in the way that the small numbers have
  // the big probablity to be generated.
  // this is for Simple8bCode
  template<typename Numeric>
  static void geometricDistribution(const float& probablity,
                              const size_t& quantity, Numeric* result) {
    // if the probablity is out of the domain, then return a empty list.
    if (probablity > 1 || probablity < 0) {
      std::cout << "Probablity should between 0 and 1." << std::endl;
    } else {
      std::default_random_engine generator;

      // generate the numbers with the given probablity.
      std::geometric_distribution<int> distribution(probablity);

      // generate the numbers according to the given quantity.
      for (size_t n = 0; n < quantity; ++n) {
        result[n] = distribution(generator);
      }
    }
  }

  // generate the numbers in the way that the small numbers have
  // the big probablity to be generated.
  // this is for others.
  template<typename Numeric>
  static void geometricDistribution(const float& probablity,
                    const size_t& quantity, vector<Numeric>& result) {
    result.reserve(quantity);
    // if the probablity is out of the domain, then return a empty list.
    if (probablity > 1 || probablity < 0) {
      std::cout << "Probablity should between 0 and 1." << std::endl;
    } else {
      std::default_random_engine generator;

      // generate the numbers with the given probablity.
      std::geometric_distribution<int> distribution(probablity);

      // generate the numbers according to the given quantity.
      for (size_t n = 0; n < quantity; ++n) {
        result.push_back(distribution(generator));
      }
    }
  }

  // generate the numbers mix uniform and geometric
  template<typename Numeric>
  static void mix(const int& min, const int& max,
                  const float& probablity, const size_t& quantity,
                  vector<Numeric>& result) {
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<> dis(min, max);

    result.reserve(quantity);
    std::default_random_engine generator;
    std::geometric_distribution<int> distribution(probablity);
    
    // generate the numbers according to the given quantity.
    for (size_t n = 0; n < quantity; ++n) {
      if ((n/240)/2 == 1) {
        result.push_back(distribution(generator));
      } else {
        result.push_back(dis(gen));
      }
    }
  }

  // generate the numbers mix uniform and geometric
  template<typename Numeric>
  static void mix(const int& min, const int& max,
                  const float& probablity, const size_t& quantity,
                  Numeric* result) {
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<> dis(min, max);
    
    std::default_random_engine generator;
    std::geometric_distribution<int> distribution(probablity);
    
    // generate the numbers according to the given quantity.
    for (size_t n = 0; n < quantity; ++n) {
      if ((n/240)/2 == 1) {
        result[n] = distribution(generator);
      } else {
        result[n] = dis(gen);
      }
    }
  }
};
#endif  // ZHIWEI_ZHANG_SOURCE_GENERATENUMBER_H_
