// Copyright 2015, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Author: Zhiwei Zhang <zhiwei.zhang@saturn.uni-freiburg.de>

#pragma once
#include <stdint.h>
#include <algorithm>
#include <assert.h>
#include "../Timer.h"

/*
 * For this class Simple8bCode there are several object methods, the 
 * prerequisite of using these object methods is making an instanz of
 * of the class Simple8bCode.
 */
namespace ad_utility_class {

  // ! Selector mask,
  // ! see: Anh & Moffat: "Index compression using 64-bit words."
  static const uint64_t SIMPLE8B_SELECTOR_MASK = 0x000000000000000F;

  // ! Selectors,
  // ! see: Anh & Moffat: "Index compression using 64-bit words."
  static const struct {
    unsigned char _itemWidth;
    unsigned char _groupSize;
    unsigned char _wastedBits;
    uint64_t _mask;
  } SIMPLE8B_SELECTORS[16] = {
    {0, 240, 60, 0x0000000000000000},  // selector 0
    {0, 120, 60, 0x0000000000000000},  // selector 1
    {1,  60,  0, 0x0000000000000001},  // selector 2
    {2,  30,  0, 0x0000000000000003},  // selector 3
    {3,  20,  0, 0x0000000000000007},  // selector 4
    {4,  15,  0, 0x000000000000000F},  // selector 5
    {5,  12,  0, 0x000000000000001F},  // selector 6
    {6,  10,  0, 0x000000000000003F},  // selector 7
    {7,   8,  4, 0x000000000000007F},  // selector 8
    {8,   7,  4, 0x00000000000000FF},  // selector 9
    {10,  6,  0, 0x00000000000003FF},  // selector 10
    {12,  5,  0, 0x0000000000000FFF},  // selector 11
    {15,  4,  0, 0x0000000000007FFF},  // selector 12
    {20,  3,  0, 0x00000000000FFFFF},  // selector 13
    {30,  2,  0, 0x000000003FFFFFFF},  // selector 14
    {60,  1,  0, 0x0FFFFFFFFFFFFFFF},  // selector 15
  };

  // ! Simple8b Compression Scheme.
  // ! See: Anh & Moffat: Index compression using 64-bit words.
  // ! Changed the following:
  // ! Selectors 0 and 1 are used to encode long streaks of the most frequent
  // ! elements. In the original paper, this is 1 since doclists are encoded.
  // ! our contexts benefit less from this.
  // ! Especially, our frequency encoding benefits from 0's much more (esp. scores)
  // ! hence we use those selectors to encode 0's instead of 1's
  class Simple8bCode {
   public:

    // defaut constructor for Simple8bCode
    Simple8bCode() {
      _encoded = new uint64_t[10];
      _cursor = 0;
      _word = 0;
      _selector = 0;
      _itemWidth = 0;
      _groupSize = 0;
    }

    // constructor with the parameters
    template<typename Numeric>
    explicit Simple8bCode(const Numeric* plaintext, const size_t& nofElements) {
      _encoded = new uint64_t[nofElements];
      encode(plaintext, nofElements, _encoded);
      _cursor = 0;
      _word = _encoded[_cursor];
      _selector = _word & SIMPLE8B_SELECTOR_MASK;
      _itemWidth = SIMPLE8B_SELECTORS[_selector]._itemWidth & 15;
      _groupSize = SIMPLE8B_SELECTORS[_selector]._groupSize & 15;
      _word = _word >> 4;
    }

    // deconstructor
    ~Simple8bCode() {
      delete[] _encoded;
    }

    // get the next word in the encoded list, which means simplely move the cursor
    // to begin of next word. Word means a 60-bits word.
    uint64_t getNextWord() {
      ++_cursor;
      _word = _encoded[_cursor];
      _selector = _word & SIMPLE8B_SELECTOR_MASK;
      _itemWidth = SIMPLE8B_SELECTORS[_selector]._itemWidth & 15;
      _groupSize = SIMPLE8B_SELECTORS[_selector]._groupSize & 15;
      _word = _word >> 4;
      return _word;
    }

    // ! the Successor function, to decode and return the next pointer.
    // ! This operation steps the file cursor (the current access point)
    // ! forwards a single pointer (or, depending on the exact context, a single integer).
    template<typename Numeric>
    void successor(Numeric* returnValue) {
      if (_groupSize == 0) {
        _word = getNextWord();
      }
      *returnValue = _word & SIMPLE8B_SELECTORS[_selector]._mask;
      _word = _word >> _itemWidth;
      _groupSize--;
    }

    // ! Forwards Search, to locate the next document pointer whose value
    // ! is greater than or equal to d, for some supplied value d.
    // ! This operation potentially shifts the file cursor to a large distance,
    // ! in terms of both pointers and stored integers.
    template<typename Numeric>
    void forwardSearch(const Numeric& reference) {
      Numeric currentNumeric = _word & SIMPLE8B_SELECTORS[_selector]._mask;
      while (currentNumeric < reference) {
        if (_groupSize <= 1) {
          _word = getNextWord();
          currentNumeric = _word & SIMPLE8B_SELECTORS[_selector]._mask;
          continue;
        }
        if (SIMPLE8B_SELECTORS[_selector]._mask < reference) {
          while (SIMPLE8B_SELECTORS[_selector]._mask < reference) {
            _word = getNextWord();
          }
          _itemWidth = SIMPLE8B_SELECTORS[_selector]._itemWidth & 15;
          _groupSize = SIMPLE8B_SELECTORS[_selector]._groupSize & 15;
          _word = _word >> 4;
          currentNumeric = _word & SIMPLE8B_SELECTORS[_selector]._mask;
          continue;
        }
        _word = _word >> _itemWidth;
        _groupSize--;
        currentNumeric = _word & SIMPLE8B_SELECTORS[_selector]._mask;
      }
    }

    // ! Forwards Seek, to shift the file cursor forwards by x integers,
    // ! for some supplied value x. In comparison with Forwards Search,
    // ! this operation steps a defined number of integers, rather than
    // ! until a particular value is located.
    void forwardSeek(const size_t& x) {
      size_t parameterofMovement = x;
      if (parameterofMovement > _groupSize) {
        while (parameterofMovement > _groupSize) {
          parameterofMovement -= _groupSize;
          ++_cursor;
          _groupSize = SIMPLE8B_SELECTORS[_encoded[_cursor]
                            & SIMPLE8B_SELECTOR_MASK]._groupSize & 15;
        }
        _word = _encoded[_cursor];
        _selector = _word & SIMPLE8B_SELECTOR_MASK;
        _itemWidth = SIMPLE8B_SELECTORS[_selector]._itemWidth & 15;
        _word = _word >> 4;
      }
      _word = _word >> (parameterofMovement * _itemWidth);
      _groupSize -= parameterofMovement;
    }

    // ! Encodes a list of elements that can be interpreted as numeric values
    // ! using the Simple8b compression scheme.
    // ! Returns the number of bytes in the encoded array, will always be a
    // ! multiple of 8.
    // ! Requires encoded to be preallocated with sufficient space.
    // ! the parameter show is used to controll the output.
    template<typename Numeric>
    static void encode(Numeric* plaintext, size_t nofElements,
                       uint64_t* encoded, const bool& show = false) {
      Timer t;
      t.start();
      // Count the number of consecutive 0's and decide if
      // selectors 0 or 1 can be used
      size_t nofElementsEncoded(0), nofCodeWordsDone(0);
      while (nofElementsEncoded < nofElements) {
        size_t itemsLeft = nofElements - nofElementsEncoded;
        // Count the number of consecutive 0's and decide if
        // selectors 0 or 1 can be used
        bool selector0(true), selector1(false);
        for (unsigned char i(0); i < std::min<size_t>(240, itemsLeft); ++i) {
          if (plaintext[nofElementsEncoded + i]) {
            selector0 = false;
            if (i >= 120) selector1 = true;
            break;
          }
        }
        if (selector0) {
          // Use selector 0 to compress 240 consecutive 0's.
          encoded[nofCodeWordsDone] = 0;
          nofElementsEncoded += 240;
          ++nofCodeWordsDone;
          continue;
        }
        if (selector1) {
          // Use selector 1 to compress 120 consecutive 0's.
          encoded[nofCodeWordsDone] = 1;
          nofElementsEncoded += 120;
          ++nofCodeWordsDone;
          // We update the itemsleft so that we can directly try to compress
          // the elements from the selector 2 not from selector 0 again.
          itemsLeft -= 120;
        }
        // Try selectors for as many items per codeword as possible,
        // take the next selector whenever it is not possible.
        for (unsigned char selector(2); selector < 16; ++selector) {
          uint64_t codeword(selector);
          unsigned char nofItemsInWord (0);
          // If an element is too large, break the loop.
          // Later we recognize that not enough elements
          // were written for the specific selector, and hence try the next
          // selector.
          while (nofItemsInWord < std::min<size_t>(
                                                   itemsLeft, SIMPLE8B_SELECTORS[selector]._groupSize)
                 && !(plaintext[nofElementsEncoded + nofItemsInWord]
                      > SIMPLE8B_SELECTORS[selector]._mask)) {
                   codeword |= (static_cast<uint64_t>(
                                                      plaintext[nofElementsEncoded + nofItemsInWord])
                                << (4 +  // Selector bits.
                                    nofItemsInWord * SIMPLE8B_SELECTORS[selector]._itemWidth));
                   ++nofItemsInWord;
                 }
          // Check if enough elements have been written to fit the selector
          // or if the loop was left earlier.
          if (nofItemsInWord == std::min<size_t>(itemsLeft,
                                                 SIMPLE8B_SELECTORS[selector]._groupSize)) {
            encoded[nofCodeWordsDone] = codeword;
            nofElementsEncoded += nofItemsInWord;
            ++nofCodeWordsDone;
            break;
          }
          // Check that the max value (60 bit) is not exceeded.
          // I put it here to same some for this check, otherwise we must
          // for each elements in hte plaintext do this check.
          assert(plaintext[nofElementsEncoded + nofItemsInWord]
                 <= 0x0FFFFFFFFFFFFFFF);
        }
      }
      t.stop();
      if (show) {
        std::cout << nofElements << "\t"
        << sizeof(uint64_t) * nofCodeWordsDone << "\t"
        << t.getUSecs() << "\t";
      }
    }
    
    
    // ! Decodes a list of elements using the Simple8b compression scheme.
    // ! Requires decoded to be preallocated with sufficient space,
    // ! i.e. sizeof(Numeric) * (nofElements + 239).
    // ! The overhead is included so that no check for noundaries
    // ! is necessary inside the decoding of a single codeword.
    template<typename Numeric>
    static void decode(uint64_t* encoded, size_t nofElements,
                       Numeric* decoded, const bool& show = false) {
      Timer t;
      t.start();
      uint64_t word;
      // Loop over full 64bit codewords and
      for (size_t nofElementsDone(0), nofCodeWordsDone(0),
           selector(encoded[nofCodeWordsDone] & SIMPLE8B_SELECTOR_MASK);
           nofElementsDone < nofElements; ++nofCodeWordsDone,
           selector = encoded[nofCodeWordsDone] & SIMPLE8B_SELECTOR_MASK) {
        word = encoded[nofCodeWordsDone] >> 4;
        for (size_t i(0); i < SIMPLE8B_SELECTORS[selector]._groupSize;
             ++i) {
          decoded[nofElementsDone++] = word & SIMPLE8B_SELECTORS[selector]._mask;
          word >>= SIMPLE8B_SELECTORS[selector]._itemWidth;
        }
      }
      t.stop();
      if (show) std::cout << t.getUSecs() << std::endl;
    }

   private:

    // the encoded code.
    uint64_t* _encoded;

    // the current word buffer.
    uint64_t _word;

    // the current selector of the word buffer in the encoded code.
    unsigned char _selector;

    // the current access position in the encoded code.
    size_t _cursor;

    // the current itemWidth of the word buffer.
    size_t _itemWidth;

    // the current number of the encoded word in the current word buffer.
    size_t _groupSize;
  };
}
